/* $Id: post.c,v 2.3 1995/04/02 11:18:02 lupus Exp lupus $ */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "post.h"
#include "clist.h"
#include "xfinans.h"
#include "post_io.h"
#include "psOutput.h"
#include "konto.h"
#include "rente.h"
#include "pante.h"
#include "aktie.h"
#include "limitPost.h"
#include "str.h"
#include "saldo.h"
#include "datoPrint.h"
#include "gregor.h"
#include "xfError.h"
#include "country.h"
#include "allocmem.h"

/* #include "reg.h" */

#define LINELEN 45

CLIST *aktPoster;                          /* linket liste af posteringer */
char *postString[MAX_POSTERINGER];  /* tekst at vise i vindue */
static int antalVises=DEFAULT_ANTAL_VISES; /* antal posteringer at vise */
static char showCommitted=0;
static char postscript=0;
static char *printer="lp";

static double currentSaldo=0.0;

void initPost(void) {
  int i;
  aktPoster=newlist(STRUCT,sizeof(struct postering)); /* create linket liste */
  for(i=0; i<MAX_POSTERINGER; i++)
    postString[i]=0;
}

int switchPS(void) {
  return (postscript = !postscript);
}

void selectPrinter(char *p) {
  printer=p;
}

void testPost(void) {
   int t;
   struct postering *p;
   printf("BEGIN\n");
   for(t=1; t<=aktPoster->size; t++) {
     p=(struct postering *)retrieve(t,aktPoster);
     printf("%s\n", p->tekst );
   }
   printf("END\n");
 }

void clearPoster(void) {
  delete_all(aktPoster); /* slet alt i den linkede liste */
}

void loadPoster(char *fname) {
  FILE *f;
  struct postering buf, *ny;

  if( (f=fopen(fname,"r"))==NULL ) {
    xfError(ERR_FILE_NOT_FOUND, NONFATAL, "file not found");
    return;
  }
  
  delete_all(aktPoster); /* slet alt i den linkede liste */

  while( load_post(f,&buf) ) {  /* indlaes post fra fil */
    insert_last((BYTE *)&buf,aktPoster); /* indsaet post i linket liste */
                                  /* saa vidt jeg forstaar laver listen sin
				     egen kopi af data, saa vi kan genbruge
				     buf uden videre -- ok, verificeret  */
  }

  fclose(f);
}

void savePoster(char *fname) {
  FILE *f;
  struct postering *p;
  int i;

  if( (f=fopen(fname,"w"))==NULL ) {
    xfError(ERR_CANT_WRITE_FILE, NONFATAL, "can't write");
    return;
  }

  for(i=0; i<aktPoster->size; i++) {
    p=(struct postering *)retrieve(i+1, aktPoster);
    save_post(f,p);
  }
  fclose(f);
}

static int startIndex(int size) {
  if( antalVises==0 ) /* antalVises=0 betyder vis alle posteringer 
			 -> altsaa start ved indx 0 */
    return 0;
  if( size-antalVises<0 )
    return 0;
  return size-antalVises;
}

void calculateCurrentSaldo(void) {  
  int i, j, saldoFirst=1;
  double saldo=0.0;
  struct postering *p;

  for(i=0; 
      i<aktPoster->size;
      i++) {
    p=(struct postering *)retrieve(i+1,aktPoster);
    if(p->dato>datoIdag() && saldoFirst) {
      saldoFirst=0;
      currentSaldo=saldo;
      currentSaldoValid(currentSaldo);
    }
    saldo+=p->bel;
  }
  if(saldoFirst) {
      saldoFirst=0;
      currentSaldo=saldo;
      currentSaldoValid(currentSaldo);
    }
}

char **almPostStrings(void) {  /* lav tekst at vise i posterings-vindue */
  char buf[256];
  int i, j, si, k, saldoFirst=1;
  double saldo=0.0, commitSaldo=0.0;
  struct postering *p;

  si=startIndex(aktPoster->size);
  for(i=0, k=0; 
      i<aktPoster->size && k<MAX_POSTERINGER-1;
      i++) {
    p=(struct postering *)retrieve(i+1,aktPoster);
    if(p->dato>datoIdag() && saldoFirst) {
      saldoFirst=0;
      currentSaldo=saldo;
      currentSaldoValid(currentSaldo);
    }
    saldo+=p->bel;
    if(showCommitted && p->committed)
      commitSaldo+=p->bel;
    if(i>=si) {
      sprintf(buf, "%d  %s", p->dato, p->tekst );
      for(j=strlen(p->tekst); j<POST_TEXT_LENGTH; j++)
	sprintf(buf+strlen(buf), " ");
      sprintf(buf+strlen(buf), "%9.2f  %9.2f", p->bel, saldo);
      if(showCommitted) {
	if(p->committed)
	  sprintf(buf+strlen(buf), "  %9.2f", commitSaldo);
	else
	  sprintf(buf+strlen(buf), "           ");
      }
      postString[k]=(char *)AllocateMemory(strlen(buf)+1);
      strcpy(postString[k], buf);
      k++;
    }
  }
  if(saldoFirst) {
      saldoFirst=0;
      currentSaldo=saldo;
      currentSaldoValid(currentSaldo);
    }
  return postString;
}

char **postStrings(void) {
  int h;

  /*printf("h=%d\n", (h & 0xffff)/15 );*/

  if(antalVises!=0) {
    h=postviewh();
    antalVises=(h & 0xffff)/15;
    if(antalVises<2)
      antalVises=DEFAULT_ANTAL_VISES;
  }

  freePostStrings();
  switch( aktKontoType() ) {
  case PANTEBREV:
    if(aktPoster->size<1)
      return postString;   /* skriv ingenting */
    return pantePostStrings((struct postering *)retrieve(1,aktPoster));
    break;
  case AKTIE:
    return aktiePostStrings(aktPoster);
    break;
  default:
    return almPostStrings();
    break;
  }
}

static void freePostStrings(void) {
  int i=0;
  while(postString[i]) {
    free(postString[i]);
    postString[i]=0;
    i++;
  }
}

void getPostAsText(int relIndx, char *dato, char *text, char *bel) {
  struct postering *p;
  int indx;
  indx=relIndx+startIndex(aktPoster->size);
  p=(struct postering *)retrieve(indx+1, aktPoster);
  sprintf(dato, "%d", p->dato);
  sprintf(text, "%s", p->tekst);
  sprintf(bel, "%.2f", p->bel);
}

int getPostCommit(int relIndx) {
  struct postering *p;
  int indx;
  indx=relIndx+startIndex(aktPoster->size);
  p=(struct postering *)retrieve(indx+1, aktPoster);
  return p->committed;
}

void okSletPost(int relIndx) {
  int indx;
  indx=relIndx+startIndex(aktPoster->size);
  delete_ith(indx+1, aktPoster);
}

void okSletPostAbsolut(int indx) {
  delete_ith(indx+1, aktPoster);
}

int okNyPost(int comm, char *dato, char *text, char *bel) {   
                          /* retur: indx paa ny post, <0 hvis fejl  */
  struct postering ny, *p;
  int i, antal, igang;
  char navn1[128], navn2[128], belbuf[50];
  double kurs;

  if( aktPoster->size >= MAX_POSTERINGER ) {
    xfError(WARN_MAX_TRANSACTIONS, NONFATAL, "max transactions");
  }

  zeroAddStrncpy(belbuf,bel,49);

  ny.committed=comm;
  ny.dato=atoi(dato);
  if( ny.dato==0 ) {
    xfError(ERR_BAD_DATE, NONFATAL, "bad date");
    return -1;
  }

  zeroAddStrncpy(ny.tekst, text, POST_TEXT_LENGTH-1);
  ny.tekst[POST_TEXT_LENGTH-1]=0;

  switch( aktKontoType() ) {
  case PANTEBREV:
    delete_first(aktPoster);  /* slet postEN, saa vi sikrer, at der kun er 1 post */
    ny.bel=atof(nocomma(belbuf));
    insert_first((BYTE *)&ny, aktPoster);
    return 0;
    break;
  case AKTIE:
    igang=0;
    navnAntalKurs(text, navn1, &antal, &kurs);
    for(i=0; i<aktPoster->size; i++) {
      p=(struct postering *)retrieve(i+1, aktPoster);
      navnAntalKurs(p->tekst, navn2, &antal, &kurs);
      if( strcmp(navn1,navn2)!=0 ) {  /* hvis navne forskellige */
	if( igang ) {
	  insert_ith(i+1,(BYTE *)&ny, aktPoster);
	  return i;
	}
      } else {
	igang=1;
	if( p->dato>ny.dato ) {
	  insert_ith(i+1,(BYTE *)&ny, aktPoster);
	  return i;
	}
      }
    }
    insert_last((BYTE *)&ny, aktPoster);
    /* testPost(); */
    return i;
    break;
  default:
    ny.bel=atof(nocomma(belbuf));
    /* printf("%d\n%s - %s\n%9.2f\n", ny.dato, ny.tekst, text, ny.bel);  */
    for(i=0; i<aktPoster->size; i++) {
      p=(struct postering *)retrieve(i+1, aktPoster);
      if( p->dato>ny.dato ) {
	insert_ith(i+1,(BYTE *)&ny, aktPoster);
	return i;
      }
    }
    insert_last((BYTE *)&ny, aktPoster);
    /* testPost(); */
    return i;
    break;
  }
}

void okRetPost(int relIndx, char *nydato, char *nytext, char *nybel) { 
  int nyIndx, glIndx, commit;

  glIndx=relIndx+startIndex(aktPoster->size);
  commit=getPostCommit(relIndx);
  nyIndx=okNyPost(commit, nydato, nytext, nybel);  /* opret ny postering */
  if(nyIndx>=0) {               /* hvis ikke fejl */
    if(nyIndx>glIndx) 
                /* hvis posten er indsat laenere nede i kontoen, end den
		   var foer */
       okSletPostAbsolut(glIndx);
    else
                /* posten er indsat laengere oppe eller samme sted som foer */
       okSletPostAbsolut(glIndx+1);
  }
}

void insertHeaderFile(FILE *to, FILE *from) {
  char buf[256];
  while( fgets(buf, 256, from) )
    fputs(buf, to);
}

void header(FILE *f, int pageNum) {
  char pbuf[128];

  setFontSize(f, "Times-ISO", 12);
  sprintf(pbuf, "%s %d", PAGE, pageNum);
  skrivXY(f, PAGE_X, PAGE_Y, pbuf);

  sprintf(pbuf, "%s %s", NUMBER, getKontoNr(aktKonto()));
  skrivXY(f, TEXT_X, PAGE_Y, pbuf);
  
  fprintf(f, "\nxfinans\n");
  setFontSize(f, "Times-ISO", 30);
  skrivOutlineXY(f, 1.1, 2.7, getKontoName(aktKonto()));
  setFontSize(f, "Times-ISO", 12);
  skrivXY(f, DATO_X, TOP_MARGIN-0.8, DATE1);
  skrivXY(f, TEXT_X, TOP_MARGIN-0.8, TEXT_LABEL);
  skrivXY(f, BEL_X, TOP_MARGIN-0.8, DEPOSIT_WITHDRAW);
  skrivXY(f, SALDO_X+0.5, TOP_MARGIN-0.8, BALANCE_LABEL);
}

void udskriv(long dato) {
  int i, j, firstLine=1;
  double saldo=0.0;
  struct postering *p;
  char buf[128];

  FILE *f, *hf;

  if(postscript) {
    if( (hf=fopen("xfin.ps", "r"))==NULL ) {
      xfError(ERR_NO_HEADER, NONFATAL, "no PS header");
      return;
    }
    f=fopen("xf.ps", "w");
    if( f==NULL ) {
      xfError(ERR_CANT_WRITE_FILE, NONFATAL, "can't write");
      return;
    }
    
    insertHeaderFile(f,hf);
    fclose(hf);
    
    initPS(f);
    /*setSize(f,11);*/
    
    for(i=0; i<aktPoster->size; i++) {
      p=(struct postering *)retrieve(i+1,aktPoster);
      saldo+=p->bel;
      
      if( p->dato>=dato ) {
	if( firstLine ) {
	  firstLine=0;
	  setFontSize(f, "Times-ISO", 11);
	  sprintf(buf, "%d", dato );
	  skriv(f, DATO_X, buf);
	  skriv(f, TEXT_X, BALANCE_LABEL);
	  sprintf(buf, "%9.2f", saldo-p->bel);  /* bel skal lige traekkes fra igen */
	  skrivRJ(f, SALDO_X+2, buf);
	  newLine(f);
	}
	setFontSize(f, "Times-ISO", 11);
	sprintf(buf, "%d", p->dato );
	skriv(f, DATO_X, buf);
	skriv(f, TEXT_X, p->tekst);
	
	sprintf(buf, "%9.2f", p->bel);
	skrivRJ(f, BEL_X+2, buf);
	
	sprintf(buf, "%9.2f", saldo);
	skrivRJ(f, SALDO_X+2, buf);
	newLine(f);
      }
    }
    if( anyLinesPrintedOnCurrentPage() )
      showpage(f);
    exitPS(f);
    fclose(f);
    sprintf(buf, "lpr -P%s xf.ps", printer);
    system(buf);
  }
  else {                    /* NOT POSTSCRIPT */
    f=fopen("xf.txt", "w");
    if( f==NULL ) {
      xfError(ERR_CANT_WRITE_FILE, NONFATAL, "can't write");
      return;
    }

    fprintf(f, "%s, %s %s\n\n", POSTING_LIST, getKontoName(aktKonto()), 
	                                  getKontoNr(aktKonto()));

    for(i=0; i<aktPoster->size; i++) {
      p=(struct postering *)retrieve(i+1,aktPoster);
      saldo+=p->bel;

      if( p->dato>=dato ) {
	if( firstLine ) {
	  firstLine=0;
	  fprintf(f, "%d %s", dato, BALANCE_LABEL);
	  for(j=strlen(BALANCE_LABEL); j<LINELEN+11; j++)
	    fprintf(f, " ");
	  fprintf(f, "%9.2f\n", saldo-p->bel);
	}
	fprintf(f, "%d %s", p->dato, p->tekst );
	for(j=strlen(p->tekst); j<LINELEN; j++)
	  fprintf(f, " ");
	fprintf(f, "%9.2f  %9.2f\n", p->bel, saldo);
      }
    }
    fclose(f);
    sprintf(buf, "lpr -P%s xf.txt", printer);
    system(buf);
  }
} 

double okRente(long d1, long d2, double f1, double f2) {
  return rente(aktPoster, d1, d2, f1, f2);
}

void defaultLimitPosts(void) {
/*  antalVises=DEFAULT_ANTAL_VISES;*/
  int h;

  h=postviewh();
  antalVises=(h & 0xffff)/15;

  if( antalVises<2 )
    antalVises=DEFAULT_ANTAL_VISES;
}

void noLimitPosts(void) {
  antalVises=0;
}

char **searchStrings(char *res[], int maxlines, char *word) {
  char buf[256], buf2[256];
  int i, j, k;
  double saldo=0.0;
  struct postering *p;

  if( word[0]=='\0' ) {
    res[0]=0;
    return res;
  }

  for(i=0, k=0; i<aktPoster->size && k<maxlines-1; i++) {
    p=(struct postering *)retrieve(i+1,aktPoster);
/*    strlower(buf2,p->tekst);
      if( regularMatch(buf2)>=0 ) { */
    if( cistrstr(p->tekst, word) ) {
      saldo+=p->bel;
      sprintf(buf, "%d  %s", p->dato, p->tekst );
      for(j=strlen(p->tekst); j<POST_TEXT_LENGTH; j++)
	sprintf(buf+strlen(buf), " ");
      sprintf(buf+strlen(buf), "%9.2f  %9.2f", p->bel, saldo);
      res[k]=(char *)AllocateMemory(strlen(buf)+1);
      strcpy(res[k], buf);
      k++;
    }
  }

  res[k]=0;
  return res;
}

int writePlotData(long fra, long til, FILE *f) {
  int i, j;
  int dage, antal=0, first=1;
  long last, thisDate=0;
  double saldo=0.0, oldsaldo=0.0, thisSaldo=0.0;
  struct postering *p;

  last=fra;

  for(i=0; i<aktPoster->size; i++) {
    p=(struct postering *)retrieve(i+1,aktPoster);
    saldo+=p->bel;
    if(last==0)
      last=p->dato;
    if( fra==0 || (fra!=0 && p->dato>=fra) ) {
      if( til==0 || (til!=0 && p->dato<=til) ) {
	if( first ) {
	  if( fra!=0 && fra!=p->dato ) {
	    antal++;
	    fprintf(f, "%d %9.2f\n", 0, oldsaldo);
	  }
	  first=0;
	} 
	thisDate=p->dato;
	thisSaldo=saldo;
	antal++;
	fprintf(f, "%d %9.2f\n", rentedage(last, p->dato), saldo);
      }
    }
    oldsaldo=saldo;
  }

  if( til>thisDate ) {
    antal++;
    fprintf(f, "%d %9.2f\n", rentedage(last, til), thisSaldo);
  }
  return antal;
}

int invertShowCommitted() {
  showCommitted=!showCommitted;
  return showCommitted;
}

void doCommit(int relIndx) {
  struct postering *p;
  int indx;
  indx=relIndx+startIndex(aktPoster->size);
  p=(struct postering *)retrieve(indx+1, aktPoster);

  p->committed=!p->committed;
  if(showCommitted)
    opdaterPostList();
  saveKonto();
}
